import unittest

from AptUrl.Helpers import parse_pkg


class Candidate:
    def __init__(self, descr):
        self.description = descr
        self.homepage = None


class MockPkg:
    def __init__(self, descr):
        self.candidate = Candidate(descr)


class HelpersTest(unittest.TestCase):
    def test_parse_pkg(self):
        pkgobj = MockPkg("summary\ndescr\n")
        self.assertEqual(parse_pkg(pkgobj), ('summary','descr\n',None))
        pkgobj = MockPkg("summary only")
        self.assertEqual(parse_pkg(pkgobj), ('summary only','',None))



# Copyright (c) 2008 Canonical
#
# AUTHOR:
# Siegfried-A. Gevatter <rainct@ubuntu.com>
#
# This file is part of AptUrl
#
# AptUrl is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# AptUrl is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with AptUrl; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import gettext
import subprocess

def _(str):
    return utf8(gettext.gettext(str))

def _n(singular, plural, n):
    return utf8(gettext.ngettext(singular, plural, n))

def utf8(str):
    if str is bytes:
        try:
            return str.decode('UTF-8')
        except UnicodeDecodeError:
            # assume latin1 as fallback
            return str.decode('latin1')
    else:
        return str

def get_dist():
    return subprocess.check_output(
        'lsb_release -c -s'.split(),
        universal_newlines=True).strip()


def parse_pkg(pkgobj):
    summary = ""
    description = ""
    pkg_description = pkgobj.candidate.description
    if pkg_description.count("\n") > 0:
        summary, description = pkg_description.split('\n', 1)
    else:
        summary = pkg_description
    lines = description.rstrip('\n').split('\n')
    if len(lines) > 1 and lines[-1].startswith('Homepage: '):
        homepage = lines[-1].split(' ', 1)[1]
        description = '\n'.join(lines[:-1])
    else:
        homepage = pkgobj.candidate.homepage
    return (summary, description, homepage)

def format_description(description):
    const = 'APTURL_DOUBLE_EMPTY_LINE_PLACEHOLDER'
    return description.replace('\n\n', const).replace('\n', ' ').replace(
        const, '\n\n')



# Copyright (c) 2007-2008 Canonical
#
# AUTHOR:
# Michael Vogt <mvo@ubuntu.com>
# With contributions by Siegfried-A. Gevatter <rainct@ubuntu.com>
#
# This file is part of AptUrl
#
# AptUrl is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# AptUrl is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with AptUrl; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os
import string
from string import Template
from .Helpers import get_dist
from .Helpers import _

class InvalidUrlException(Exception):
    def __init__(self, url, msg=""):
        self.url = url
        self.message = msg
    def __str__(self):
        return self.message

MAX_URL_LEN=255

# substituion mapping
apturl_substitution_mapping = {
    "distro" : get_dist(),
    "kernel" : os.uname()[2] 
}

# whitelist for the uri
whitelist = []
whitelist.extend(string.ascii_letters)
whitelist.extend(string.digits)
whitelist.extend(['_',':','?','/','+','.','~','=','<','>','-',',','$','&'])


class AptUrl(object):
    " a class that contains the parsed data from an apt url "
    def __init__(self):
        self.package = None
        self.schema = None
        self.minver = None
        self.refresh = None
        # for added repos
        self.keyfile = None
        self.repo_url = None
        self.dist = '/'     
        # for known sections 
        self.section = []
        # for known channels
        self.channel = None

def is_format_package_name(string):
    " return True if string would be an acceptable name for a Debian package "
    
    return (string.replace("+", "").replace("-", "").replace(".", "").replace(":", "").isalnum() 
        and string.islower() and string[0].isalnum() and len(string) > 1)

def do_apt_url_substitution(apt_url, mapping):
    " substitute known templates against the field package and channel "
    for field in ["package","channel"]:
        if getattr(apt_url, field):
            s=Template(getattr(apt_url, field))
            setattr(apt_url, field, s.substitute(mapping))

def match_against_whitelist(raw_url):
    " test if the url matches the internal whitelist "
    for char in raw_url:
        if not char in whitelist:
            raise InvalidUrlException(
                raw_url, _("Non whitelist char in the uri"))
    return True

def set_value(apt_url, s):
    " set a key,value pair from string s to AptUrl object "
    (key, value) = s.split("=")
    try:
        if ' ' in value:
            raise InvalidUrlException(apt_url, _("Whitespace in key=value"))
        if type(getattr(apt_url, key)) == type([]):
            getattr(apt_url, key).append(value)
        else:
            setattr(apt_url, key, value)
    except Exception as e:
        raise InvalidUrlException(apt_url, _("Exception '%s'") % e)


def parse(full_url, mapping=apturl_substitution_mapping):
    " parse an apt url and return a list of AptUrl objects "
    # apt:pkg1?k11=v11?k12=v12,pkg2?k21=v21?k22=v22,...
    res = []
    
    if len(full_url) > MAX_URL_LEN:
        url = "%s ..." % full_url[0:(MAX_URL_LEN // 10)]
        raise InvalidUrlException(url, _("Url string '%s' too long") % url)

    # check against whitelist
    match_against_whitelist(full_url)
    for url in full_url.split(";"):
        if not ":" in url:
            raise InvalidUrlException(url, _("No ':' in the uri"))

        # now parse it

        (schema, packages) = url.split(":", 1)
        packages = packages.split(",")

        for package in packages:
            apt_url = AptUrl()
            apt_url.schema = schema
            # check for schemas of the form: apt+http://
            if schema.startswith("apt+"):
                apt_url.repo_url = schema[len("apt+"):] + ":" + package.split("?",1)[0]
            else:
                if "?" in package:
                    apt_url.package = package.split("?")[0].lstrip("/")
                else:
                    apt_url.package = package.lstrip("/")

            # now parse the ?... bits
            if "?" in package:
                key_value_pairs = package.split("?")[1:]
                for s in key_value_pairs:
                    if "&" in s:
                        and_key_value_pairs = s.split("&")
                        for s in and_key_value_pairs:
                            set_value(apt_url, s)
                    else:
                        set_value(apt_url, s)

            # do substitution (if needed) 
            do_apt_url_substitution(apt_url, mapping)
            
            # check if the package name is valid
            if not is_format_package_name(apt_url.package):
                raise InvalidUrlException(url, "Invalid package name '%s'" % apt_url.package)
            
            res.append(apt_url)
    return res    



import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GObject
GObject.threads_init()

import os
import sys
import apt_pkg
import subprocess
import tempfile

from AptUrl.UI import AbstractUI
from AptUrl import Helpers
from AptUrl.Helpers import _

APTURL_UI_FILE = os.environ.get(
    # Set this envar to use a test .ui file.
    'APTURL_UI_FILE',
    # System file to use if the envar is not set.
    '/usr/share/apturl/apturl-gtk.ui'
    )


class GtkUI(AbstractUI):
    def __init__(self):
        Gtk.init_check(sys.argv)
        # create empty dialog
        self.dia_xml = Gtk.Builder()
        self.dia_xml.set_translation_domain("apturl")
        self.dia_xml.add_from_file(APTURL_UI_FILE)
        self.dia = self.dia_xml.get_object('confirmation_dialog')
        self.dia.start_available = lambda: Gtk.main_quit()
        self.dia.start_error = lambda: Gtk.main_quit()
        self.dia.exit = lambda: Gtk.main_quit()
        self.dia.realize()
        self.require_update = False

    # generic dialogs
    def _get_dialog(self, dialog_type, summary, msg="",
                    buttons=Gtk.ButtonsType.CLOSE):
        " internal helper for dialog construction "
        d = Gtk.MessageDialog(parent=self.dia,
                              flags=Gtk.DialogFlags.MODAL,
                              type=dialog_type,
                              buttons=buttons)
        d.set_title("")
        d.set_markup("<big><b>%s</b></big>\n\n%s" % (summary, msg))
        d.set_icon(Gtk.IconTheme.get_default().load_icon('package-x-generic', 16, False))
        d.set_keep_above(True)
        d.realize()
        d.get_window().set_functions(Gdk.WMFunction.MOVE)
        return d

    def error(self, summary, msg=""):
        d = self._get_dialog(Gtk.MessageType.ERROR, summary, msg)
        d.run()
        d.destroy()
        return False

    def message(self, summary, msg="", title=""):
        d = self._get_dialog(Gtk.MessageType.INFO, summary, msg)
        d.set_title(title)
        d.run()
        d.destroy()
        return True

    def yesNoQuestion(self, summary, msg, title="", default='no'):
        d = self._get_dialog(Gtk.MessageType.QUESTION, summary, msg,
                             buttons=Gtk.ButtonsType.YES_NO)
        d.set_title(title)
        res = d.run()
        d.destroy()
        if res != Gtk.ResponseType.YES:
            return False
        return True

    def askInstallPackage(self, package, summary, description, homepage):
        # populate the dialog
        dia = self.dia
        dia_xml = self.dia_xml
        header = _("Install additional software?")
        body = _("Do you want to install package '%s'?") % package
        dia.set_title('')
        header_label = dia_xml.get_object('header_label')
        header_label.set_markup("<b><big>%s</big></b>" % header)
        body_label = dia_xml.get_object('body_label')
        body_label.set_label(body)
        description_text_view = dia_xml.get_object('description_text_view')
        tbuf = Gtk.TextBuffer()
        desc = "%s\n\n%s" % (summary, Helpers.format_description(description))
        tbuf.set_text(desc)
        description_text_view.set_buffer(tbuf)
        dia.set_icon(Gtk.IconTheme.get_default().load_icon('package-x-generic', 16, False))

        # check if another package manager is already running
        # FIXME: just checking for the existance of the file is
        #        not sufficient, it need to be tested if it can
        #        be locked via apt_pkg.get_lock()
        #        - but that needs to run as root
        #        - a dbus helper might be the best answer here
        #args = (update_button_status, dia_xml.get_object("yes_button"),
        #    dia_xml.get_object("infolabel"))
        #args[0](*args[1:])
        #timer_id = GObject.timeout_add(750, *args )

        # show the dialog
        res = dia.run()
        #GObject.source_remove(timer_id)
        if res != Gtk.ResponseType.YES:
            dia.hide()
            return False

        return True

    # progress etc
    def doUpdate(self):
        self.require_update = True
        win = self.dia.get_window()
        try:
            xid = win.get_xid()
        except Exception as e:
            print(e)
            xid = 0
        FNULL = open(os.devnull, 'w')
        cmd = ["sudo", "/usr/lib/linuxmint/mintUpdate/checkAPT.py", "--use-synaptic", "%d" % xid]
        comnd = subprocess.Popen(' '.join(cmd), stdout=FNULL, stderr=subprocess.STDOUT, shell=True)
        returnCode = comnd.wait()

    def doInstall(self, apturl):
        self.dia.hide()
        win = self.dia.get_window()
        try:
            xid = win.get_xid()
        except Exception as e:
            print(e)
            xid = 0

        cmd = ["pkexec", "/usr/sbin/synaptic", "--hide-main-window",  "--non-interactive", "--parent-window-id", "%s" % xid]
        cmd.append("-o")
        cmd.append("Synaptic::closeZvt=true")
        f = tempfile.NamedTemporaryFile()
        for pkg in [apturl.package]:
            pkg_line = "%s\tinstall\n" % pkg
            f.write(pkg_line.encode("utf-8"))
        cmd.append("--set-selections-file")
        cmd.append("%s" % f.name)
        f.flush()
        comnd = subprocess.Popen(' '.join(cmd), shell=True)
        returnCode = comnd.wait()
        f.close()
        self.dia.exit()

if __name__ == "__main__":
    ui = GtkUI()
    ui.error("foo","bar")
